#!/usr/bin/env python
#
# FSL wrapper functions - named fslwrap so as not to clash with any potential
# 'official' FSL python modules which would probably be named 'fsl'
#
# TODO
#
# FSL context? For given path, workdir
# 'temp=True' for temp run outside workdir
# Or 'temp' vs 'persist' contexts?

import os, sys
import glob
import shutil
import shlex
import subprocess
import math
import errno
import tempfile

import nibabel as nib
import numpy as np

# Where to look for 'local' programs - default to dir of calling program
# This enables users to override standard FSL programs by putting their
# own copies in the same directory as the script they are calling
LOCAL_DIR = os.path.dirname(os.path.abspath(sys.argv[0]))
#print("FSLWRAP: using local binaries dir: %s" % LOCAL_DIR)

def set_localdir(localdir):
    global LOCAL_DIR
    LOCAL_DIR = localdir

class Workspace:
    """
    A workspace for FSL processing

    Based on a working directory, which may be temporary or persistent.
    """

    def __init__(self, workdir=None, log=sys.stdout, imgs=[], debug=False, echo=False):
        """
        Create workspace

        :param workdir: If specified, use this path for the working directory. Will be created
        if not already existing
        :param log: File stream to write log output to (default: sys.stdout)
        :param imgs: List of Image object which will be save to working directory
        """
        if workdir is not None:
            self.workdir = workdir
            mkdir(workdir)
            self.is_temp = False
        else:
            self.workdir = tempdir("fsl")
            self.is_temp = True
        
        for img in imgs:
            self.add_img(img)

        self.log = log
        self.debug = debug
        self.echo = echo

        self.mcflirt = self._runnable(mcflirt)
        self.bet = self._runnable(bet)
        self.flirt = self._runnable(flirt)
        self.fast = self._runnable(fast)
        self.apply_xfm = self._runnable(apply_xfm)
        
    def __del__(self):
        if self.is_temp:
            shutil.rmtree(self.workdir)

    def add_img(self, img):
        img.save(os.path.join(self.workdir, img.fname))

    def sub(self, name, imgs=[]):
        """
        Create a sub-workspace, (i.e. a subdir of this workspace)

        This inherits the log output from the parent workspace

        :param name: Name of subdir
        :imgs imgs: Images to copy to the sub workspace
        """
        return Workspace(os.path.join(self.workdir, name), log=self.log, imgs=imgs)

    def run(self, fslprog, *args, **kwargs):
        """
        Run an FSL program

        This is used to run FSL programs for which wrappers have not yet been provided

        :param fslprog: Name of program to run
        """
        return Prog(fslprog)(*args, workdir=self.workdir, log=self.log, debug=self.debug, echo=self.echo, **kwargs)

    def write_file(self, text, name):
        """
        Write a text file to the workspace

        :param text: Text to put in the file
        :param name: File name
        """
        with open(os.path.join(self.workdir, name), "w") as f:
            f.write(text)

    def _runnable(self, fn):
        def func(*args, **kwargs):
            return fn(*args, workdir=self.workdir, log=self.log, debug=self.debug, echo=self.echo, **kwargs)
        return func

class Prog:
    """ 
    An FSL program

    The program is checked for in the 'local dir' first, then FSLDEVDIR/bin, then FSLDIR/bin.
    This search is performed at execution time, so changes to these directories can be 
    made in between calls.

    An FSL program is run in a 'working directory'. If no working directory is given, a temporary
    working directory is created and deleted afterwards. FSL programs work on image and other data
    stored in files and write output to other files.

    If required, a set of Image objects can be pre-written to the working directory using the
    input_imgs argument.

    If required, a set of output image filenames can be given in the output_imgs argument, which 
    will be read back in from the working directory and returned as Image objects.

    If required, a set of output non-image filenames can be given in the output_files argument, 
    which will be ready back in from the working directory and returned as string objects.

    Hence, you can basically use fsl.Prog in more-or-less two distinct ways.
      1. The traditional way, specify a real working directory, output is saved there and you
         can read it back in to your program as required. output_imgs and output_files are not
         generally given.

      2. The non-working-dir way, where you do not specify workdir and instead ask for image
         and non-image data to be returned as Image objects and strings. This is required 
         because the temporary files they were in will be deleted after execution. Your program
         can do with these as it wishes, including writing them out to other files.

    It's also possible to combine (1) and (2), e.g. files are written to a real working directory as 
    a record, but also retrieved back into the program for convenience. It's also possible to run
    some FSL commands (e.g. ones which produce a lot of output, most of it not required) in temporary
    working directories, and just pull out the required data to write to the real working directory.
    """

    def __init__(self, cmd):
        self.cmd = cmd

    def echo(self, *args, **kwargs):
        if self._echo:
            self.log.write(*args, **kwargs)
            self.log.write("\n")

    def debug(self, *args, **kwargs):
        if self._debug:
            self.log.write(*args, **kwargs)
            self.log.write("\n")

    def _find(self):
        """ 
        Find the program, either in the 'local' directory, or in $FSLDEVDIR/bin or $FSLDIR/bin 
        This is called each time the program is run so the caller can control where programs
        are searched for at any time
        """
        dirs = [
            LOCAL_DIR,
            os.path.join(os.environ.get("FSLDEVDIR", LOCAL_DIR), "bin"),
            os.path.join(os.environ.get("FSLDIR", LOCAL_DIR), "bin"), 
        ]

        for d in dirs:
            ex = os.path.join(d, self.cmd)
            if os.path.isfile(ex) and os.access(ex, os.X_OK):
                return ex
        
        return self.cmd

    def __call__(self, argline="", input_imgs=[], output_imgs=[], output_files=[], **kwargs):
        """
        Run FSL command

        Returns tuple of: 
          (output images (list of Image), output files (list of strings), command stdout)
        """

        self._debug = kwargs.pop("debug", False)
        self._echo = kwargs.pop("echo", False)
        self.log = kwargs.pop("log", sys.stdout)
        workdir = kwargs.pop("workdir", None)

        # Create working directory if required. If not specified use temp directory and
        # delete afterwards
        if workdir is None:
            workdir = tempdir(self.cmd)
            delete_workdir = True
        else:
            workdir = mkdir(workdir, fail_if_exists=False, warn_if_exists=False)
            delete_workdir = False
    
        cwd = os.getcwd()
        try:
            os.chdir(workdir)

            # Write input images to files in working dir
            for img in input_imgs:
                self.debug("Saving to %s in %s\n" % (img.fname, os.getcwd()))
                img.save(img.fname)

            cmd = self._find()
            cmd_args = shlex.split(cmd + " " + argline)
            for arg, value in kwargs.items():
                cmd_args.append(arg)
                if value != "": cmd_args.append(value)

            cmd_stdout = ""
            self.echo(" ".join(cmd_args))
            p = subprocess.Popen(cmd_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            while 1:
                retcode = p.poll() # returns None while subprocess is running
                cmd_stdout += p.stdout.readline()
                if retcode is not None: break
            if retcode != 0:
                raise RuntimeError(cmd_stdout)
                
            output_imgs = [Image(fname, load_data=True) for fname in output_imgs]
            output_files = [self._read_text_file(fname) for fname in output_files]

            if delete_workdir:
                shutil.rmtree(workdir)

            self.echo(cmd_stdout)
            return output_imgs, output_files, cmd_stdout
        finally:
            os.chdir(cwd)

    def _read_text_file(self, fpath):
        with open(fpath, "r") as f:
            return f.read()

class Image: 
    """
    Named image data

    An image is identified by a *name* which should be unique within the context.
    It will be used as the basis for a file name with an added Nifti extension.

    If an image is loaded from a file it's name will be taken from the filename, 
    minus the directory path and Nifti extension. 

    If an image is created from existing Numpy data, a name must be provided

    Attributes:
      ``iname`` - Image name. Interpreted as file name minus extension
      ``fname`` - File name with Nifti extension but no directory
      ``fpath`` - Full file path with Nifti extension and directory
      ``ipath`` - Image path with directory but no extension
      ``ext``   - Extension
      ``dpath`` - Full path to directory
      ``role``  - Role this image plays in the context (e.g. 'calibration', 't1')
    """
    def __init__(self, name,  data=None, role="Image", **kwargs):
        if name is None:
            raise RuntimeError("%s image - no name given" % role)
        self.role = role
        self._data = None
        self.ext_matches = []
        self.default_ext = kwargs.pop("default_ext", ".nii.gz")
        self.set_name(name)

        if data is not None:
            # We have been given the data - create a Nifti image from it
            # Can be based on a previous Image to inherit affine, header etc.
            if "base" in kwargs:
                self.nii = nib.Nifti1Image(data, None, kwargs.pop("base").nii.header)
                self.nii.update_header()
            else:
                self.nii = nib.Nifti1Image(data, np.identity(4))
            self.shape = data.shape
            self.ndim = data.ndim

            # Might want to set the file immediately
            if kwargs.pop("save", False):
                self.save()
            
        else:
            self.load()
            # Might want to get the data immediately if it is from a temp file
            if kwargs.pop("load_data", False): 
                self.data()
        
    def load(self, name=None):
        """
        Load the image

        Only the metadata is actually loaded immediately - image data is only
        loaded on calling ``data()``

        :param name: If specified, will call ``set_name`` first with this name.
            Otherwise, existing image file name/path is used to locate the file
        """
        if name is not None:
            self.set_name(name)

        if len(self.ext_matches) > 1:
            raise RuntimeError("%s: file %s is ambiguous" % (self.role, self.fname))
    
        # Load the file
        try:
            self.nii = nib.load(self.fpath)
        except:
            raise RuntimeError("%s: file %s not found" % (self.role, self.fname))
        self.shape = self.nii.shape
        self.ndim = len(self.shape)
        return self

    def save(self, name=None):
        """
        Save the image to a file

        :param name: If specified, will call ``set_name`` first with this name.
        """
        if name is not None:
            self.set_name(name)
        self.nii.to_filename(self.fpath)
        return self

    def set_name(self, name):
        """
        Set the name of the image

        :param name: New name for the image

        The name of the image can be used just as an identifier (if we
        are just using ``Image`` as a container for Numpy data) but
        is also interpreted as a file name if we ``load`` or ``save``
        the image. If ``name`` contains a directory path and/or extension
        these are stripped off and used to populate the file attributes
        of the image.

        The following attributes are set by this method:
        
            iname
            fname
            ipath
            fpath
            dpath

        This method will not save the image to a file. To do that, use ``save``
        """
        self.dpath, self.fname = os.path.split(name)
        
        # Keep relative paths relative
        if self.dpath != "": self.dpath = os.path.abspath(self.dpath)

        if "." in self.fname:
            self.iname, self.ext = self.fname.split(".", 1)
            self.ipath = os.path.join(self.dpath, self.iname)
        else:
            self.iname = self.fname
            self.ipath = os.path.join(self.dpath, self.iname)
            self._guess_extension()
            self.fname = self.fname + self.ext

        self.fpath = os.path.join(self.dpath, self.fname)
        return self

    def _guess_extension(self):
        """
        Guess the extension of the file, if the name was given without extension
        
        Looks for already existing files with the same ipath and a Nifti extension.
        The attribute ``ext_matches`` stores the matching extensions. If no matches
        are found, the default Nifti extension is set (``.nii.gz``)
        """
        exts = ["", ".nii", ".nii.gz"]
        self.ext_matches = []
        for ext in exts:
            if os.path.exists("%s%s" % (self.ipath, ext)):
                self.ext_matches.append(ext)

        if len(self.ext_matches) == 0:
            self.ext = self.default_ext
        else:
            self.ext = self.ext_matches[0]

    def data(self):
        """
        Return the image data

        If image was initialized from a file and has not yet been loaded, this
        will load the data from the file.

        :returns: A Numpy array of image data
        """
        if self._data is None:
            self._data = self.nii.get_data()
        return self._data

    def summary(self, log=sys.stdout):
        """
        Output a readable summary of the image

        :param log: File or output stream to write to (default: sys.stdout)
        """
        log.write("%s: %s\n" % (self.role.ljust(30), self.iname))
        log.write("Full path                     : %s\n" % self.fpath)

    def check_shape(self, shape):
        """
        Check that image shape matches another shape

        :param shape: Shape to check against
        """
        if len(self.shape) != len(shape):
            raise RuntimeError("%s: expected %i dims, got %i" % (self.file_type, len(shape), len(self.shape)))
        if self.shape != shape:
            raise RuntimeError("%s: shape (%s) does not match (%s)" % (self.file_type, str(self.shape), str(shape)))

    def derived(self, data, name=None, suffix=None, **kwargs):
        """
        Create a derived image based on this one, but with different data

        Useful because it preserves orientation info, etc
        May be overridden in subclasses to preserve subclass-specific attributes

        :param data: Numpy data for derived image
        :param name: Name for new image (can be simple name or full filename)
        :param suffix: If name not specified, construct by adding suffix to original image name

        Any further keyword parameters are passed to the Image constructor
        """
        if name is None and suffix is None:
            name = self.ipath
        elif name is None:
            name = self.ipath + suffix
        return Image(name, data=data, base=self, **kwargs)

maths = Prog("fslmaths")
roi = Prog("fslroi")
stats = Prog("fslstats")
merge = Prog("fslmerge")

def fast(img, **kwargs):
    output_name, extra_args = _get_std(img, "_fast", kwargs)
    args = "-o %s %s %s" % (output_name, extra_args, img.iname)
    imgs, files, stdout = Prog("fast")(
        args, input_imgs=[img],output_imgs=[output_name + "_seg"], **kwargs
    )
    return imgs[0]

def bet(img, **kwargs):
    output_name, extra_args = _get_std(img, "_bet", kwargs)
    args = "%s %s -m %s" % (img.iname, output_name, extra_args)
    imgs, files, stdout = Prog("bet")(
        args, input_imgs=[img], output_imgs=[output_name, output_name + "_mask"], **kwargs
    )
    return tuple(imgs)

def flirt(img, ref, **kwargs):
    output_name, extra_args = _get_std(img, "_reg", kwargs)
    args = "-in %s -ref %s -out %s %s" % (img.iname, ref.iname, output_name, extra_args)

    files = []
    output_mat = kwargs.pop("output_mat", None)
    output_invmat = kwargs.pop("output_invmat", None)
    if output_mat is not None or output_invmat is not None:
        args += " -omat %s" % output_mat
        files.append(output_mat)

    imgs, files, stdout = Prog("flirt")(
        args, input_imgs=[img], output_imgs=[output_name], output_files=files, **kwargs
    )

    ret = [imgs[0], ]
    if output_mat is not None:
        ret.append(text_to_matrix(files[0]))
    if output_invmat is not None:
        invmat = np.linalg.inv(text_to_matrix(files[0]))
        ret.append(invmat)
    
    return tuple(ret)
    
def apply_xfm(img, ref, xfm, **kwargs):
    output_name, extra_args = _get_std(img, "_reg", kwargs)
    args = "-in %s -ref %s -applyxfm -init %s -out %s %s" % (img.iname, ref.iname, xfm, output_name, extra_args)
    imgs, files, stdout = Prog("flirt")(
        args, input_imgs=[img], output_imgs=[output_name], **kwargs
    )
    return imgs[0]

def mcflirt(img, cost=None, ref=None, **kwargs):
    output_name, extra_args = _get_std(img, "_mc", kwargs)
    args = "-in %s -out %s %s" % (img.iname, output_name, extra_args)
    if cost is not None:
        args += " -cost %s" % cost
    if ref is not None:
        args += " -r %s" % ref.iname
    imgs, files, stdout = Prog("mcflirt")(
        args, input_imgs=[img], output_imgs=[output_name], **kwargs
    )
    return imgs[0]

def _get_std(img, suffix, kwargs):
    return kwargs.pop("output_name", img.iname + suffix), kwargs.pop("extra_args", "")

def matrix_to_text(mat):
    rows = []
    for row in mat:
        rows.append(" ".join([str(v) for v in row]))
    return "\n".join(rows)

def text_to_matrix(text):
    fvals = []
    ncols = -1
    lines = text.splitlines()
    for line in lines:
        # Discard comments
        line = line.split("#", 1)[0].strip()
        # Split by commas or spaces
        vals = line.replace(",", " ").split()
        # Ignore empty lines
        if len(vals) == 0: continue
        # Check correct number of columns
        if ncols < 0: ncols = len(vals)
        elif len(vals) != ncols:
            raise RuntimeError("File must contain a matrix of numbers with fixed size (rows/columns)")
        # Check all data is numeric
        for val in vals:
            try:
                fval = float(val)
            except:
                raise RuntimeError("Non-numeric value '%s' found in file %s" % (val, filename))
        fvals.append([float(v) for v in vals])     
    return np.array(fvals)

def imcp(src, dest):
    Prog("imcp")("%s %s" % (src, dest))
    
def mkdir(dirname, fail_if_exists=False, warn_if_exists=True):
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno == errno.EEXIST:
            if fail_if_exists: raise
            elif warn_if_exists: print("WARNING: mkdir - Directory %s already exists" % dirname)
    return os.path.abspath(dirname)

def tempdir(suffix, debug=False):
    if debug:
        tmpdir = os.path.join(os.getcwd(), "tmp_%s" % suffix)
        mkdir(tmpdir)
    else:
        tmpdir = tempfile.mkdtemp("_%s" % suffix)
    print("Using temporary dir: %s" % tmpdir)
    return tmpdir